
package com.wutka.util;

import java.util.*;

/** ResourcePool holds a common set of resources that are given out
    in first-in-first-out order. The pool assumes that all resources
    are equally weighted. Since some resource pools may need to allocate
    additional resources when no more are available, the getOverFlowResource
    method can be overridden to create new ones.
    <P>
    When the resource is allocated, the requesting thread can associate an
    object with the allocated resource. One way to make use of this feature is
    to store the name of the method that is requesting the resource. When
    debugging code, you may want to look at the allocated resources and what
    methods have allocated them. Someties you can find methods that aren't
    releasing resources after allocating them.
    <P>
    If you allocate resources within a try block, you should release them
    in the finally section of the block. That way, the resource is released
    no matter what exception is thrown.


    @author Mark Wutka
    @version 1.0
*/

public class ResourcePool
{

/** The resources that are currently free */
    protected Vector availableResources;

/** Resources that have been allocated out of the pool */
    protected Vector usedResources;

/** User-defined info about a resource that has been allocated */
    protected Vector usedResourceInfo;

/** Flag to make sure at least one thread has received notification */
    boolean receivedWakeup;

/** The number of threads waiting for notification */
    int numThreadsWaiting;

/** Creates a new resource pool */
    public ResourcePool()
    {
        availableResources = new Vector();
        usedResources = new Vector();
        usedResourceInfo = new Vector();
        receivedWakeup = true;
        numThreadsWaiting = 0;
    }

/** Adds another resource to the pool
 * @param resource The resource being added
 */
    public synchronized void addResource(Object resource)
    {
 // Assume the resource is free, add it to the free list
        availableResources.addElement(resource);
    }

/** Removes a resource from the pool
 * @param resource The resource to be removed
 */
    public synchronized void removeResource(Object resource)
    {
// Remove the resource from the free list
        availableResources.removeElement(resource);

// See if the resources is in the used list
        int pos = usedResources.indexOf(resource);

// If the resource is in the used list, remove it and its associated info
        if (pos >= 0) {
            usedResources.removeElementAt(pos);
            usedResourceInfo.removeElementAt(pos);
        }
    }

/** Allocates a resource when the pool is empty. By default, this method
 *    returns null, indicating that the requesting thread must wait. This
 *    allows a thread pool to expand when necessary, allowing for spikes in
 *    activity.
 *    @return A new resource, or null to force the requester to wait
 */
    protected synchronized Object getOverflowResource()
    {
        return null;
    }

/** Requests a resource from the pool, waiting forever if one is not available.
 * No extra information is associated with the allocated resource.
 * @return The allocated resource
 */
    public Object getResource()
    {
        return getResource(null, -1);
    }

/** Requests a resource from the pool, waiting forever if one is not available.
 * @param allocationInfo A value associated with the allocated resource
 * @return The allocated resource
 */
    public Object getResource(Object allocationInfo)
    {
        return getResource(allocationInfo, -1);
    }

/** Requests a resource from the pool, waiting forever if one is not available.
 * @param timeout The maximum amount of time (in milliseconds)
 * to wait for the resource
 * @return The allocated resource
 */
    public Object getResource(long timeout)
    {
        return getResource(null, timeout);
    }

/** Grabs a resource from the free list and moves it to the used list.
 * This method is really the core of the resource pool. The rest of the class
 * deals with synchronization around this method.
 * @param allocationInfo The info to be associated with the allocated resource
 * @return The allocated resource
 */
    protected synchronized Object getResourceFromList(Object allocationInfo)
    {
// See if there is a resource available.
        if (availableResources.size() > 0) {

// Get the first resource from the free list
            Object resource = availableResources.elementAt(0);

// Remove the resource from the free list
            availableResources.removeElement(resource);

// Add the resource and its associated info to the used list
            usedResources.addElement(resource);
            usedResourceInfo.addElement(allocationInfo);

            return resource;
        }

        return null;
    }

/** Performs a wait for a specified number of milliseconds.
 * @param timeout The number of milliseconds to wait
 * (wait forever if timeout < 0)
 */
    protected synchronized void doWait(long timeout)
    {
        try {
            if (timeout < 0) {
                wait();
            }
            else {
                wait(timeout);
            }
        }
        catch (Exception ignore) {
        }
    }

/** Requests a resource from the pool, waiting forever if one is not available.
 * @param allocationInfo A value associated with the allocated resource
 * @param timeout The maximum amount of time (in milliseconds)
 * to wait for the resource
 * @return The allocated resource
 */
    public Object getResource(Object allocationInfo, long timeout)
    {
 // See if there is a resource in the pool already
        Object resource = getResourceFromList(allocationInfo);
        if (resource != null)
        {
            return resource;
        }

// Figure out when to stop waiting
        long endTime = System.currentTimeMillis() + timeout;

        do {

            synchronized(this) {
// See if there are any available resources in the pool
                if (availableResources.size() == 0) {

// Allow subclasses to provide overflow resources
                    resource = getOverflowResource();

// If there was a resource allocated for overflow, add it to the used list
                    if (resource != null)
                    {
                        usedResources.addElement(resource);
                        usedResourceInfo.addElement(allocationInfo);

                        return resource;
                    }
                }
            }

// Wait for a resource to be allocated

// Figure out the longest time to wait before timing out
            long maxWait = endTime - System.currentTimeMillis();
            if (timeout < 0) maxWait = -1;

// Indicate that there is a thread waiting for a wakeup
            numThreadsWaiting++;

// Wait for a wakeup
            doWait(maxWait);

            numThreadsWaiting--;

// Only mention the received wakeup if the timeout hasn't expired
            if ((timeout < 0) || (System.currentTimeMillis() < maxWait)) {
                receivedWakeup = true;
            }

// See if there is now a resource in the free pool
            resource = getResourceFromList(allocationInfo);
            if (resource != null)
            {
                return resource;
            }

// Keep looping while the timeout hasn't expired (loop forever if there is
// no timeout.
        } while ((timeout < 0) || (System.currentTimeMillis() < endTime));

        return null;
    }

/** Releases a resource back to the pool of available resources
 * @param resource The resource to be returned to the pool
 */
    public void releaseResource(Object resource)
    {
        int pos = -1;

        synchronized(this) {
// Make sure the resource is in the used list
            pos = usedResources.indexOf(resource);

// If the resource was in the used list, remove it from the used list and
// add it back to the free list
            if (pos >= 0) {
                usedResources.removeElementAt(pos);
                usedResourceInfo.removeElementAt(pos);
                availableResources.addElement(resource);
            }
        }

// If we released a resource, wake up any threads that may be waiting
        if (pos >= 0)
        {
            doWakeup();
        }
    }

/** Performs a notifyAll (which requires a synchronized method) */
    protected synchronized void doNotify()
    {
        try {
            notifyAll();
        }
        catch (Exception ignore) {
        }
    }

    protected void doWakeup()
    {
// Wake up any threads waiting for the resource
        receivedWakeup = false;
        do {
            try {
                   doNotify();
            }
            catch (Exception ignore) {
            }
        }
// Keep looping while there are threads waiting and none have received a wakeup
        while ((numThreadsWaiting > 0) && !receivedWakeup);
    }

/** Return a copy of the resources in the free list
 * @return The resources in the free list
 */
    public synchronized Vector getAvailableResources()
    {
        return (Vector) availableResources.clone();
    }

/** Return a copy of the resources in the used list
 * @return The resources in the used list
 */
    public synchronized Vector getUsedResources()
    {
        return (Vector) usedResources.clone();
    }

/** Return a copy of the associated info for the used resources
 * @return The info associated with the allocated resources
 */
    public synchronized Vector getUsedResourceInfo()
    {
        return (Vector) usedResourceInfo.clone();
    }
}
